package be.iminds.ilabt.jfed.highlevel.util;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecInterface;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.stream.Collectors;

/**
 * ESpecMetaFileCreator
 */
public class ESpecMetaFileCreator {
    private final Experiment experiment;
    private final GeniUser geniUser;

    private final static ObjectMapper MAPPER = new ObjectMapper();

    public ESpecMetaFileCreator(Experiment experiment, GeniUser geniUser) {
        this.experiment = experiment;
        this.geniUser = geniUser;
    }

    public byte[] getManifestXml() {
        if (experiment.getSlice().getManifestRspec() == null)
            throw new IllegalStateException("There is no manifest while uploading manifest.xml to node");
        String manifestString = experiment.getSlice().getManifestRspec().getRspecXmlString();
        if (manifestString == null)
            throw new IllegalStateException("There is no manifest string while uploading manifest.xml to node");
        return manifestString.getBytes(StandardCharsets.UTF_8);
    }

    public byte[] getExperimentInfoJson() {
        if (experiment.getSlice().getManifestRspec() == null)
            throw new IllegalStateException("There is no manifest while uploading manifest.xml to node");
        ModelRspec modelRspec = experiment.getSlice().getManifestRspec().getModelRspec(ModelRspecType.IMMUTABLE);
        if (modelRspec == null)
            throw new IllegalStateException("There is no parsable manifest while generating meta experiment-info.json");

        Map<String, Object> res = new HashMap<>();
        Map<String, Map<String, ?>> nodes = new HashMap<>();
        res.put("nodes", nodes);

        for (RspecNode node : modelRspec.getNodes()) {
            if (node.getClientId() == null)
                continue;

            Map<String, Object> nodeInfo = new HashMap<>();
            nodes.put(node.getClientId(), nodeInfo);

            nodeInfo.put("client_id", node.getClientId());
            if (node.getAnsibleGroups() == null || node.getAnsibleGroups().isEmpty()) {
                nodeInfo.put("ansible_groups", Collections.emptyList());
            } else {
                nodeInfo.put("ansible_groups", new ArrayList<>(node.getAnsibleGroups()));
            }
            if (node.getInterfaces() == null || node.getInterfaces().isEmpty()) {
                nodeInfo.put("internal_ips", Collections.emptyList());
                nodeInfo.put("connected_nodes", Collections.emptyList());
            } else {
                List<String> ips = node.getInterfaces().stream()
                        .map(RspecInterface::getIpAddresses)
                        .filter(Objects::nonNull)
                        .flatMap(Collection::stream)
                        .map(RspecInterface.IpAddress::getAddress)
                        .filter(Objects::nonNull)
                        .collect(Collectors.toList());
                nodeInfo.put("internal_ips", ips);

                List<Map<String, Object>> connectedNodes = new ArrayList<>();
                nodeInfo.put("connected_nodes", connectedNodes);
                for (RspecInterface iface : node.getInterfaces()) {
                    List<? extends RspecInterface> otherIfaces = new ArrayList<>(iface.getLink().getInterfaces());
                    otherIfaces.remove(iface);
                    for (RspecInterface otherIface : otherIfaces) {
                        Map<String, Object> conNodeInfo = new HashMap<>();
                        conNodeInfo.put("node", otherIface.getNode().getClientId());
                        conNodeInfo.put("local_iface", iface.getClientId());
                        conNodeInfo.put("remote_iface", otherIface.getClientId());
                        conNodeInfo.put("local_ips", iface.getIpAddresses() == null ?
                                Collections.emptyList() :
                                iface.getIpAddresses().stream()
                                        .filter(Objects::nonNull)
                                        .map(RspecInterface.IpAddress::getAddress)
                                        .filter(Objects::nonNull)
                                        .collect(Collectors.toList()));
                        conNodeInfo.put("remote_ips", otherIface.getIpAddresses() == null ?
                                Collections.emptyList() :
                                otherIface.getIpAddresses().stream()
                                        .filter(Objects::nonNull)
                                        .map(RspecInterface.IpAddress::getAddress)
                                        .filter(Objects::nonNull)
                                        .collect(Collectors.toList()));
                        connectedNodes.add(conNodeInfo);
                    }
                }
            }
            if (node.getLoginServices() == null || node.getLoginServices().isEmpty()) {
                nodeInfo.put("ssh_login", Collections.emptyList());
            } else {
                List<Map> sshLogins = node.getLoginServices().stream()
                        .filter(Objects::nonNull)
                        .filter(ls -> Objects.equals(ls.getAuthentication(), "ssh-keys"))
                        .map(ls -> {
                            Map<String, Object> map = new HashMap<>();
                            map.put("hostname", ls.getHostname());
                            map.put("username", ls.getUsername());
                            map.put("port", ls.getIntPort());
                            map.put("hostandport", ls.getHostname()+":"+ls.getPort());
                            map.put("all", ls.getUsername()+"@"+ls.getHostname()+":"+ls.getPort());
                            return map;
                        })
                        .collect(Collectors.toList());
                nodeInfo.put("ssh_login", sshLogins);
            }
        }

        Map<String, String> user = new HashMap<>();
        List<Map<String, ?>> sshUsers = new ArrayList<>();
        Map<String, String> slice = new HashMap<>();
        Map<String, String> project = new HashMap<>();
        res.put("user", user);
        res.put("ssh_users", sshUsers);
        res.put("slice", slice);
        res.put("project", project);

        GeniUrn sliceUrn = experiment.getSlice().getUrn();
        slice.put("name", experiment.getSlice().getName());
        slice.put("urn", experiment.getSlice().getUrnString());

        GeniUrn projectUrn = GeniUrn.createGeniUrnFromEncodedParts(
                sliceUrn.getEncodedTopLevelAuthority(),
                "project",
                experiment.getSlice().getSubAuthorityName()
        );
        project.put("name", experiment.getSlice().getSubAuthorityName());
        project.put("urn", projectUrn.getValue());

        user.put("name", geniUser.getUserUrn().getResourceName());
        user.put("urn", geniUser.getUserUrnString());

        if (experiment.getUserSpecs() != null) {
            for (UserSpec userSpec : experiment.getUserSpecs()) {
                Map<String, Object> userSpecInfo = new HashMap<>();
                sshUsers.add(userSpecInfo);
                userSpecInfo.put("urn", userSpec.getUrn());
                userSpecInfo.put("ssh_keys", userSpec.getSshKey());
                GeniUrn urn = GeniUrn.parse(userSpec.getUrn());
                if (urn != null) userSpecInfo.put("name", urn.getResourceName());
            }
        }

        try {
            return MAPPER.writerWithDefaultPrettyPrinter().writeValueAsBytes(res);
        } catch (JsonProcessingException e) {
            throw new RuntimeException("Problem generating JSON while generating meta experiment-info.json", e);
        }
    }
}
